// 2017-00-10 BiQuadr.cpp  
#include <math.h>
#include "poly34.h"		    // solution of cubic and quartic equation
#include "BiQuadr.h"		//  


static double f_min012(double fm, double f0, double f1, double &x);

//-----------------------------------------------------------------------------
//   f(x)  f(x) = a0 + a1*x + a2*x*x , :
// f(-1) = fm,
// f( 0) = f0,
// f( 1) = f1.
//      [-1,1],     x,  f(x)
// a0 = f0,
// a1 = (f1-fm)/2,
// a2 = (f1-2*f0+fm)/2 = (f1+fm)/2 - f0,
// x  = (fm-f1)/(2*(f1-2*f0+fm)).
static double f_min012(double fm, double f0, double f1, double &x)
{
	double a2 = 0.5*(f1 + fm - 2 * f0);
	if (a2 <= 0) {						//  ,   
		if (fm < f1) {
			x = -1;
			return fm;
		}
		else {
			x = 1;
			return f1;
		}
	}
	// now a2 > 0
	double a1 = 0.5*(f1 - fm);
	double a0 = f0;
	x = -0.5*a1 / a2;
	if (x<-1) {
		x = -1;
		return fm;
	}
	if (x>1) {
		x = 1;
		return f1;
	}
	return a0 + a1*x + a2*x*x;
}

//-----------------------------------------------------------------------------
//     9- :
// (-1,-1) ( 0,-1) ( 1,-1)     f[0]   f[1]   f[2]
// (-1, 0) ( 0, 0) ( 1, 0) ->  f[3]   f[4]   f[5]
// (-1, 1) ( 0, 1) ( 1, 1)     f[6]   f[7]   f[8]
//
//    
//
//   f(x,y) = a[0] + a[1]*x + a[2]*x*x + (a[3] + a[4]*x + a[5]*x*x)*y  + (a[6] + a[7]*x + a[8]*x*x)*y*y
//
//     a0 = f[4]
// 2 * a1 = -f[3] + f[5]
// 2 * a2 = f[3] - 2 * f[4] + f[5]
// 4 * a3 = -2 * f[1] + 2 * f[7]
// 4 * a4 = f[0] - f[2] - f[6] + f[8]
// 4 * a5 = -f[0] + 2 * f[1] - f[2] + f[6] - 2 * f[7] + f[8]
// 4 * a6 = 2 * f[1] - 4 * f[4] + 2 * f[7]
// 4 * a7 = -f[0] + f[2] + 2 * f[3] - 2 * f[5] - f[6] + f[8]
// 4 * a8 = f[0] - 2 * f[1] + f[2] - 2 * f[3] + 4 * f[4] - 2 * f[5] + f[6] - 2 * f[7] + f[8]
//
void findBiQuad(double *f, double *a)	//     a[i]  9-  f[j]
{
	a[0] = f[4];
	a[1] = 0.5*(-f[3] + f[5]);
	a[2] = 0.5*(f[3] - 2 * f[4] + f[5]);
	a[3] = 0.25*(-2 * f[1] + 2 * f[7]);
	a[4] = 0.25*(f[0] - f[2] - f[6] + f[8]);
	a[5] = 0.25*(-f[0] + 2 * f[1] - f[2] + f[6] - 2 * f[7] + f[8]);
	a[6] = 0.25*(2 * f[1] - 4 * f[4] + 2 * f[7]);
	a[7] = 0.25*(-f[0] + f[2] + 2 * f[3] - 2 * f[5] - f[6] + f[8]);
	a[8] = 0.25*(f[0] - 2 * f[1] + f[2] - 2 * f[3] + 4 * f[4] - 2 * f[5] + f[6] - 2 * f[7] + f[8]);
}

//-----------------------------------------------------------------------------
// yMin = -1/2*(a3+a4*x+a5*x^2)/(a6+a7*x+a8*x^2)
// eqv  = 2*a8*(4*a2*a8-a5^2)*x^5
//      + (-2*a4*a5*a8+4*a1*a8^2-3*a5^2*a7+16*a2*a7*a8)*x^4
//      + 4*(2*a1*a7*a8-a4*a5*a7-a5^2*a6+4*a2*a6*a8+2*a2*a7^2)*x^3
//      + (4*a1*a7^2+8*a1*a6*a8+2*a3*a4*a8-2*a3*a5*a7-6*a4*a5*a6+16*a2*a6*a7-a4^2*a7)*x^2
//      + 2*(4*a1*a6*a7+4*a2*a6^2+a3^2*a8-2*a3*a5*a6-a4^2*a6)*x
//      + a3^2*a7-2*a3*a4*a6+4*a1*a6^2
// type[i] -  ,
// type[i] = -1 - ,
// type[i] =  0 - ,
// type[i] = +1 - ,

//
int extremBiQuad(double *a, int *type, double *x0, double *y0, double *z0)	//    a[i]    (x0[i],y0[i]),    (z0[i])    
{
	double co[6];
	co[0] = 2 * a[8] * (4 * a[2] * a[8] - a[5] * a[5]);																	// x^5
	co[1] = -2 * a[4] * a[5] * a[8] + 4 * a[1] * a[8] * a[8] - 3 * a[5] * a[5] * a[7] + 16 * a[2] * a[7] * a[8];	// x^4
	co[2] = 4 * (2 * a[1] * a[7] * a[8] - a[4] * a[5] * a[7] - a[5] * a[5] * a[6] + 4 * a[2] * a[6] * a[8] + 2 * a[2] * a[7] * a[7]);	// x^3
	co[3] = 4 * a[1] * a[7] * a[7] + 8 * a[1] * a[6] * a[8] + 2 * a[3] * a[4] * a[8] - 2 * a[3] * a[5] * a[7] - 6 * a[4] * a[5] * a[6] + 16 * a[2] * a[6] * a[7] - a[4] * a[4] * a[7];	//x^2
	co[4] = 2 * (4 * a[1] * a[6] * a[7] + 4 * a[2] * a[6] * a[6] + a[3] * a[3] * a[8] - 2 * a[3] * a[5] * a[6] - a[4] * a[4] * a[6]);	// x^1
	co[5] = a[3] * a[3] * a[7] - 2 * a[3] * a[4] * a[6] + 4 * a[1] * a[6] * a[6];									// x^0
	int res;				//  
	if (fabs(co[0]) > 1e-12) { //  5- 
		double et = 1 / co[0];
		res = SolveP5(x0, et*co[1], et*co[2], et*co[3], et*co[4], et*co[5]);
	}
	else if (fabs(co[1]) > 1e-12) { //  4- 
		double et = 1 / co[1];
		res = SolveP4(x0, et*co[2], et*co[3], et*co[4], et*co[5]);
	}
	else if (fabs(co[2]) > 1e-12) { //  3- 
		double et = 1 / co[2];
		res = SolveP3(x0, et*co[3], et*co[4], et*co[5]);
	}
	else if (fabs(co[3]) > 1e-12) { //  2- 
		double et = 1 / co[3];
		res = SolveP2(x0, et*co[4], et*co[5]);
	}
	else if (fabs(co[4]) > 1e-12) { //  1- 
		double et = -1 / co[4];
		res = 1;
		x0[0] = et*co[5];
	}
	else {							//  0- 
		res = 0;
	}
	// res =  
	for (int i = 0; i < res; i++) {
		double x = x0[i];
		double x2 = x*x;
		double y = -0.5*(a[3] + a[4] * x + a[5] * x2) / (a[6] + a[7] * x + a[8] * x2);
		double y2 = y*y;
		y0[i] = y;
		z0[i] = a[0] + a[1] * x + a[2] * x2 + (a[3] + a[4] * x + a[5] * x2)*y + (a[6] + a[7] * x + a[8] * x2)*y2;
		//   :
		double fxx = 2 * (a[2] + y*a[5] + y*y * a[8]);
		double fxy = a[4] + 2*(x*a[5] + y*(a[7] + 2*x*a[8]));
		double fyy = 2 * (a[6] + x*a[7] + x*x * a[8]);
		double D = fxx*fyy - fxy*fxy;
		type[i] = D<=0? 0:  fxx > 0 ? -1 : +1;
	}
	return res;
}
//-----------------------------------------------------------------------------
double vBiSQ(double *a, double x, double y)	//  9 - . a[i].     (x,y)
{
	return a[0] + a[1] * x + a[2] * x*x + (a[3] + a[4] * x + a[5] * x*x)*y + (a[6] + a[7] * x + a[8] * x*x)*y*y;
}
//-----------------------------------------------------------------------------
double minBiSq(double *f, double &x, double &y, double maxOffset)	//    f  9- . (x,y):=   
{
	double a[9];			// .  
	findBiQuad(f, a);
	int type[5];
	double x0[5], y0[5], z[5];
	int K = extremBiQuad(a, type, x0, y0, z); //  
	double re = 1e300;
	x = y = 0;
	//printf("--------K=%d\n", K);
	for (int i = 0; i < K; i++) {
		//printf(" %c %10.3f: <%10.5f,%10.5f>\n", type[i]<0 ? '-' : type[i]>0 ? '+' : ' ', z[i], x0[i], y0[i]);
		if (z[i] < re && fabs(x0[i])<maxOffset && fabs(y0[i])<maxOffset) {
			re = z[i];
			x = x0[i];
			y = y0[i];
		}
	}
	//printf("--------\n");
	// ,       9- 
	double *ft = f;
	for (int y1 = -1; y1 <= 1; y1++)
		for (int x1 = -1; x1 <= 1; x1++) {
			if (*ft < re) {
				re = *ft;
				x = x1;
				y = y1;
			}
			ft++;
		}
	// ,      
	double f4, x4;
	// 
	f4 = f_min012(f[0], f[1], f[2], x4);
	if (f4 < re) {
		re = f4;
		x = x4;
		y = -1;
	}
	// 
	f4 = f_min012(f[6], f[7], f[8], x4);
	if (f4 < re) {
		re = f4;
		x = x4;
		y = 1;
	}
	// 
	f4 = f_min012(f[0], f[3], f[6], x4);
	if (f4 < re) {
		re = f4;
		x = -1;
		y = x4;
	}
	// 
	f4 = f_min012(f[2], f[5], f[8], x4);
	if (f4 < re) {
		re = f4;
		x =  1;
		y = x4;
	}
	return re;
}
//-----------------------------------------------------------------------------
