// forders.cpp: 2016-05-18 Frobenius orders(c,a,b,p)
// (c) Khashin S.I.  khash2@gmail.com

//  ,   ,     
//      z=a+b*sqrt(c)     p,
//    2    .

//      unsigned *ford    ,  ,
//   ford[i] = (p*p-1)/FrobOrder(a,b,c,p).
//    >=2^32,     0   
//  z64.

#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif

#ifdef __MINGW32__
  #define __USE_MINGW_ANSI_STDIO 1
#endif


#include <stdio.h>

#include "z64.h"			// sieve of Eratosthenes v.3
#include "forders.h"		// FrobOrder(a,b,c,p)

//----------------------------------------------------------------------------
int forders::a=2;
int forders::b=1;
int forders::c=-1;
unsigned  forders::nord=0;
unsigned *forders::ord=NULL;
u64       forders::maxP_=0;
//----------------------------------------------------------------------------
int forders::calc(int a0, int b0, int c0, u64 maxN)// calculate seive and orders(base)
{
	a=a0; b=b0; c=c0;
	if( z64::maxN()<maxN ) return -1;
	if( nord>0 ) delete [] ord;
	nord = (unsigned) z64::Pi(maxN);
	ord = new unsigned[nord];
	ord[0]=0;							// because p(0)=0
	u64 p, q;
	i64 a1, b1, c1;
	c1 = c;
	for( unsigned i=1; i<nord; i++ )
	{
		if( i%1000==0 )//|| i>43000 )
			//printf("%7u/%7u K\r", i, nord );
			printf("%7u/%7u K\r", i/1000, nord/1000 );
		p = z64::p(i);
		a1 = a;	while( a1<0 ) a1 += p;
		b1 = b; while( b1<0 ) b1 += p;
		u64 fo = z64::FrobOrder(a1, b1, c1, p);
		if( fo==0 ) ord[i]=0;
		else {
			q = (p*p-1)/z64::FrobOrder(a1, b1, c1, p);
			ord[i] = q<=MAX32 ? (unsigned)q : 0;
		}
	}
	maxP_ = p;
	return 0;
}
//----------------------------------------------------------------------------
u64 forders::fordI(unsigned i)			// FrobOrder(a,b,c, p(i))
{
	u64 p = z64::p(i);
	if( i<nord ) {
		unsigned q= ord[i];
		if( q>0 ) return (p*p-1)/q;
	}
	i64 a1, b1;
	a1 = a;	while( a1<0 ) a1 += p;
	b1 = b; while( b1<0 ) b1 += p;
	return z64::FrobOrder(a1, b1, c, p);
}
u64 forders::fordP(u64 p0)			// FrobOrder(a,b,c, p0)
{
	if( !z64::isPrime(p0) ) return 0;
	unsigned ip0 = (unsigned)z64::Pi(p0);
	if( ip0<nord ) {
		unsigned q= ord[ip0];
		if(q>0) return (p0*p0-1)/q;
	}

	i64 a1, b1;
	a1 = a;	while( a1<0 ) a1 += p0;
	b1 = b; while( b1<0 ) b1 += p0;
	return z64::FrobOrder(a1, b1, c, p0);
}
//----------------------------------------------------------------------------
unsigned forders::fCordI(unsigned i)			// co-FrobOrder(a,b,c, p(i))
{
	return i < nord ? ord[i] : 0;
}
u64 forders::fCordP(u64 p0)			// FrobOrder(a,b,c, p0)
{
	if (!z64::isPrime(p0)) return 0;
	unsigned ip0 = (unsigned)z64::Pi(p0);
	if (ip0 < nord) return ord[ip0];

	i64 a1, b1;
	a1 = a;	while (a1<0) a1 += p0;
	b1 = b; while (b1<0) b1 += p0;
	return (p0*p0-1)/z64::FrobOrder(a1, b1, c, p0);
}
//----------------------------------------------------------------------------
static char fname_buf[256];
char *forders::forder_file_name( int a0, int b0, int c0){
	//sprintf(fname_buf, "c:/a/primes/fords/ford_%04d_%04d_%04d.dat", c0, a0, b0);
	sprintf(fname_buf, "%sfords\\ford_%04d_%04d_%04d.dat", z64::path, c0, a0, b0);	//   
	return fname_buf;
}
//----------------------------------------------------------------------------
int forders::save(char *fname)	// save orders
{
	FILE *f=fopen(fname, "wb");
	if( !f ) { fclose(f); return 1; }
	unsigned label = 777777777;
	fwrite( &label, sizeof(unsigned), 1, f );
	fwrite( &a,     sizeof(int)     , 1, f );
	fwrite( &b,     sizeof(int)     , 1, f );
	fwrite( &c,     sizeof(int)     , 1, f );
	fwrite( &nord,  sizeof(unsigned), 1, f );
	fwrite( ord, 4, nord, f );
	fclose(f);
	return 0;
}
//----------------------------------------------------------------------------
int  forders::load(char *fname, u64 N)	// load orders (base)
{
	//printf(" D01) orders::load(%s, %llu)\n", fname, N);
	if( nord>0) { delete ord; nord=0; }
	FILE *f=fopen(fname, "rb");
	if( !f ) { return 1; }
	unsigned label;
	size_t res;
	res = fread( &label, sizeof(unsigned), 1, f );
	if( res != 1 ) return 2;
	if( label != 777777777 ) return 3;

	res = fread( &a, sizeof(int), 1, f );
	if( res != 1 ) return 2;
	if( a <-100000 || a>100000 ) return 4;

	res = fread( &b, sizeof(int), 1, f );
	if( res != 1 ) return 2;
	if( b <-100000 || b>100000 ) return 4;

	res = fread( &c, sizeof(int), 1, f );
	if( res != 1 ) return 2;
	if( c <-100000 || c>100000 ) return 4;

	res = fread( &nord, sizeof(unsigned), 1, f );
	if( res != 1 ) return 2;
	if( N>0 )
	{
		unsigned iN = (unsigned) z64::Pi(N);
		if( nord>iN ) nord = iN;				//    ,  
	}
	//printf(" D02) nord=%u\n", nord);
	ord = new unsigned[nord];
	res = fread( ord, sizeof(unsigned), nord, f );
	//printf(" D03) res =%u\n", res);
	fclose(f);
	maxP_ = z64::p(nord-1);
	//printf(" D04) nord-1=%u, maxP_=%llu\n", nord-1, maxP_);
	return res == nord? 0 : 5;
}
//----------------------------------------------------------------------------
int  forders::fill(int a0, int b0, int c0, u64 maxN)	// load seive and orders(2)
{
	char *fname = forder_file_name(a0,b0,c0);
	if( load(fname , maxN)==0 ) return 0;
	//printf(" 3 \n");
	calc(a0, b0, c0, maxN);
	save(fname);
	return 0;
}
//----------------------------------------------------------------------------
