// u128.h : 16-bytes numbers
//
#pragma warning(disable : 4996)

#include <stdio.h>
#include <stdlib.h>		// for lldiv
#include <intrin.h>

#include "u128.h"			// 16-bytes numbers

const u64 MAXu64 = 0xffffffffffffffffLL;
const u64 uK = 1000, uM = uK*uK, uG = uK*uM;


 bool even(u64 x) { return (x & 1) == 0; }
 bool odd(u64 x) { return (x & 1) == 1; }

static char u128_buf[64];
static u64  pow2m1[64] = {
	0   , 1    ,  3  ,	7   , 0xf  , 0x1f  , 0x3f  , 0x7f  ,
	0xff, 0x1ff,0x3ff, 0x7ff, 0xfff, 0x1fff, 0x3fff, 0x7fff,
	0xffff, 0x1ffff,0x3ffff, 0x7ffff, 0xfffff, 0x1fffff, 0x3fffff, 0x7fffff,
	0xffffff, 0x1ffffff,0x3ffffff, 0x7ffffff, 0xfffffff, 0x1fffffff, 0x3fffffff, 0x7fffffff,
	0xffffffffLL, 0x1ffffffffLL,0x3ffffffffLL, 0x7ffffffffLL, 0xfffffffffLL, 0x1fffffffffLL, 0x3fffffffffLL, 0x7fffffffffLL,
	0xffffffffffLL, 0x1ffffffffffLL,0x3ffffffffffLL, 0x7ffffffffffLL, 0xfffffffffffLL, 0x1fffffffffffLL, 0x3fffffffffffLL, 0x7fffffffffffLL,
	0xffffffffffffLL, 0x1ffffffffffffLL,0x3ffffffffffffLL, 0x7ffffffffffffLL, 0xfffffffffffffLL, 0x1fffffffffffffLL, 0x3fffffffffffffLL, 0x7fffffffffffffLL,
	0xffffffffffffffLL, 0x1ffffffffffffffLL,0x3ffffffffffffffLL, 0x7ffffffffffffffLL, 0xfffffffffffffffLL, 0x1fffffffffffffffLL, 0x3fffffffffffffffLL, 0x7fffffffffffffffLL,
};

// ariphmetics ----------------------------------------------------------------
bool u128::odd() {		// this += a_
	return a&1;
}
void u128::add(u64 a_) {		// this += a_
	b += _addcarry_u64(0, a, a_, &a);
}
void u128::add(u128 &f) {		// this += f
	unsigned char c = _addcarry_u64(0, a, f.a, &a);
	_addcarry_u64(c, b, f.b, &b);
}
void u128::mul(u64 a_) {		// this *= a_
	u64 t;
	if(b)
	b = _umul128(b, a_, &t);
	a = _umul128(a, a_, &t);
	b += t;
}
void u128::mul(u128 &f) {		// this *= f
	if (f.b == 0) { mul(f.a); return; }
	// (a + M*b)*(f.a + M*f.b) = a*f.a + M*(b*f.a + a*f.b)
	u64 t, abba;
	abba = _umul128(b, f.a, &t) + _umul128(a, f.b, &t);
	a = _umul128(a, f.a, &t);
	b = abba+t;
}
int u128::shift2s() {		// divide by 2 till odd
	int res = 0;
	if (a == 0) { 
		if (b == 0) { return 0; }
		a = b; b = 0; res += 32; }
	//  a != 0;
	unsigned long ind;
	_BitScanForward64(&ind, a);
	if (ind == 0) return res;
	res += ind;
	a = __shiftright128(a, b, (unsigned char)ind);
	b >>= ind;
	return res;
}

int u128::ooStep() {						// one step of (3n+1)-problem, from odd to odd, returns the number of division by 2
	u128 t(*this);
	add(*this);						// this *=2
	add(t);							// this = this0*3
	add(1);							// this = this0*3+1
	return shift2s();				// this = (this0*3+1)/2^s
}

int u128::log2() {						// high bit index 0..127
	unsigned long ind;
	if (b != 0) {
		_BitScanReverse64(&ind, b);
		return 64 + (int)ind;
	}
	_BitScanReverse64(&ind, a);
	return ind;
}

// operators ----------------------------------------------
u128 u128::operator=(u64 a_) {
	a = a_; b = 0;
	return *this;
}
u128 u128::operator=(u128 &f) {
	a = f.a; b = f.b;
	return *this;
}
u128 u128::operator+(u64 a_) {
	u128 t(*this);
	t.add(a_);
	return t;
}
u128 u128::operator+(u128 &f) {
	u128 t(*this);
	t.add(f);
	return t;
}
u128 u128::operator++() {
	add(1);
	return *this;
}
u128 u128::operator+=(u64 a_) {
	add(a_);
	return *this;
}
u128 u128::operator+=(u128 &f) {
	add(f);
	return *this;
}
u128 u128::operator<<=(unsigned char x) {		// multipyly by 2^x
	b = __shiftleft128(a, b, x);
	a <<= x;
	return *this;
}
u128 u128::operator>>=(unsigned char x) {		// divide by 2^x
	a = __shiftright128(a, b, x);
	b >>= x;
	return *this;
}
// power -------------------------------------------------
void u128::pow2(int n) {							// this = 2^n
	if (n < 0 || n >= 128) { a = b = 0; return; }
	if (n < 64) { a = 1LL << n; b = 0; return; }
	a = 0; b = 1LL << (n - 64);
}
// power -------------------------------------------------
void u128::pow(u64 f, int n) {					// this = f^n
	if (n <= 0) { a = 1; b = 0; return; }
	a = f; b = 0;
	while (n-- > 1) mul(f);
}
// subtract ---------------------------------------------
u128 u128::operator--() {
	if (a > 0) {
		a--; return *this;
	}
	a = MAXu64;
	b--;
	return *this;
}
u128 u128::operator-=(u64 a_) {
	if (a >= a_) {
		a-=a_; return *this;
	}
	a -= a_;
	b--;
	return *this;
}
u128 u128::operator-=(u128 &f) {
	if (a >= f.a) {
		a -= f.a;
	}
	else {
		a -= f.a;
		b--;
	}
	b-=f.b;
	return *this;
}
u128 u128::operator-(u64 a_) {
	u128 t(*this);
	t-=a_;
	return t;
}
u128 u128::operator-(u128 &f) {
	u128 t(*this);
	t -= f;
	return t;
}
// compare u128 ----------------------------------------
bool u128::operator==(u128 &f) {
	return a == f.a && b == f.b;
}
bool u128::operator!=(u128 &f) {
	return a != f.a || b != f.b;
}
bool u128::operator<(u128 &f) {
	if (b < f.b) return true;
	if (b > f.b) return false;
	// now b == f.b
	return a < f.a;
}
bool u128::operator>(u128 &f) {
	if (b > f.b) return true;
	if (b < f.b) return false;
	// now b == f.b
	return a > f.a;
}
bool u128::operator<=(u128 &f) {
	if (b < f.b) return true;
	if (b > f.b) return false;
	// now b == f.b
	return a <= f.a;
}
bool u128::operator>=(u128 &f) {
	if (b > f.b) return true;
	if (b < f.b) return false;
	// now b == f.b
	return a >= f.a;
}
// compare u64----------------------------------------
bool u128::operator==(u64 a_) {
	return  b == 0 && a == a_;
}
bool u128::operator!=(u64 a_) {
	return  b != 0 || a != a_;
}
bool u128::operator<(u64 a_) {
	if (b > 0) return false;
	//  b == 0
	return a < a_;
}
bool u128::operator>(u64 a_) {
	if (b > 0) return true;
	//  b == 0
	return a > a_;
}
bool u128::operator<=(u64 a_) {
	if (b > 0) return false;
	//  b == 0
	return a <= a_;
}
bool u128::operator>=(u64 a_) {
	if (b > 0) return true;
	//  b == 0
	return a >= a_;
}
// division by 10-----------------------------------------------
int u128::div10() {			// this /=10, returns remainder
		const u64 M1 = 1844674407370955161LL; // 2^64 = 6 + 10*M1
		u128 t(*this);
		u64 ar, br;
		ar = a % 10; a = a / 10;
		br = b % 10; b = b / 10;
		this->add(M1*br);
		ar += br * 6;
		this->add(ar / 10);
		ar = ar % 10;
		return (int) ar;
}
void u128::div(u128 &f, u128 &d, u128&r)	// d = this/f, r = this%f   !!slowly!!
{
	if (*this < f) { d = 0; r = *this; return; }
	// now this >=f
	r = *this;
	u128 ff=f;								// r/f
	d = 0;
	int deg2 = log2() - f.log2();			// deg2 >= 0
	u128 p2; p2.pow2(deg2);
	ff <<= deg2;
	do {
		if (r>=ff) { 
			d += p2;
			r -= ff;
		}
		p2 >>= 1;
		ff >>= 1;
		deg2--;
	} while (deg2>=0);
}
void u128::div(u64 &f_, u128 &d, u128&r)	// d = this/f, r = this%f !!slowly!!
{
	u128 f(f_);
	div(f, d, r);
}
// division --------------------------------------------------------------------
u128 u128::operator/=(u64 a_)
{
	u128 d, r;
	div(a_, d, r);
	*this = d;
	return *this;
}
u128 u128::operator/=(u128 &f)
{
	u128 d, r;
	div(f, d, r);
	*this = d;
	return *this;
}
u128 u128::operator/(u64 a_)
{
	u128 d, r;
	div(a_, d, r);
	return d;
}
u128 u128::operator/(u128 &f)
{
	u128 d, r;
	div(f, d, r);
	return d;
}
u128 u128::operator%=(u64 a_)
{
	u128 d, r;
	div(a_, d, r);
	*this = r;
	return r;
}
u128 u128::operator%=(u128 &f)
{
	u128 d, r;
	div(f, d, r);
	*this = r;
	return r;
}
u128 u128::operator%(u64 a_)
{
	u128 d, r;
	div(a_, d, r);
	return r;
}
u128 u128::operator%(u128 &f)
{
	u128 d, r;
	div(f, d, r);
	return r;
}
// string ---------------------------------------------
char *u128::toString() {
		if (b == 0) {
			sprintf(u128_buf, "%llu", a);
			return u128_buf;
		}
		int i;
		for (i = 0; i < 64; i++) u128_buf[i] = 0;
		u128 x(*this);
		i = 40;
		while (x > 0) {
			int r = x.div10();
			u128_buf[i--] = '0' + r;
		}
		for (int j = i + 1; j <= 40; j++)
			u128_buf[j - i - 1] = u128_buf[j];
		u128_buf[40 - i] = 0;

		return u128_buf;
	}
char *u128::toStringH() {		// hex
		if (b == 0)
			sprintf(u128_buf, "0x%llx", a);
		else
			sprintf(u128_buf, "0x%llx%016llx", b, a);
		return u128_buf;
	}

//-----------------------------------------------------------------------------

